'' /////////////////////////////////////////////////////////////////////////////
'' Graphics Renderer for Donkey Kong Remake
'' AUTHOR: Steve Waddicor
'  Adapted from:
'  *****************************
'  *  TV Driver v1.0           *
'  *  (C) 2004 Parallax, Inc.  *
'  *****************************
'' //////////////////////////////////////////////////////////////////////////////

CON
  param_count   = 10
  sprites       = 32

VAR
  long  cog
  long  renderer_ptr_block

PUB start(renderer_ptr) : okay

'' Start renderer - starts a cog
'' returns false if no cog available

  renderer_ptr_block := renderer_ptr
  okay := (cog := cognew(@entry,renderer_ptr)) > 0

DAT

'******************************
'* Assembly language renderer *
'******************************

                        org
'
'
' Entry
'
entry
                        ' Load parameters
                        mov     ptr,PAR                 
                        movd    :load,#renderer_params
                        mov     count,#param_count
:load                   rdlong  renderer_params,ptr
                        add     :load,inc_dest
                        add     ptr,#4
                        djnz    count,#:load

                        or      dira,#1
                        mov     current_line,rend_cog_number  'start rendering from the top

do_line
                        call    #debug_start
                        ' Load sprites
                        mov     ptr,PAR
                        add     ptr,#param_count*4           
                        movd    :load,#rend_sprites
                        mov     count,#sprites
:load                   rdlong  0-0,ptr
                        add     :load,inc_dest
                        add     ptr,#4
                        djnz    count,#:load

                        movd    scan_line_store1,#cog_scanline_buffer
                        movd    scan_line_store2,#cog_scanline_buffer+1

                        ' Init debug 
                        mov     ptr,PAR
                        add     ptr,#param_count*4+32*4 
                        rdlong  rend_debugging,ptr wc,wz
        if_e            mov     hub_map_ptr,rend_map    'point to first tile (upper-leftmost)
        if_ne           mov     hub_map_ptr,rend_debugging
        
                        mov     tile_row,current_line   'then move down to current line first tile
                        shr     tile_row,#3             '8 lines per tile
                        shl     tile_row,#5             '32 tiles per row
                        add     hub_map_ptr,tile_row

                        mov     tile_line,current_line  'line offset within tile
                        and     tile_line,#$07          '8 lines per tile
                        shl     tile_line,#1            '2 bytes per line
                        mov     tile_table,rend_tiletable  'get a local copy of tiletable address
                        add     tile_table,tile_line    'offset it to read correct tile_line

                        mov     x,rend_ht               'set horizontal tiles

do_tile                 rdbyte  tile,hub_map_ptr        'read tile
                        shl     tile,#4                 '16 bytes per tile
                        add     tile,tile_table         'convert to pointer
                        rdword  pixels1,tile            'Pixels 1 = %%xxxxxxxx01234567          16 bits

                        mov     pixels2,pixels1         'Pixels 2 = %%xxxxxxxx01234567          16 bits
                        shr     pixels1,#8              'Pixels 1 = %%xxxxxxxxxxxx0123           8 bits

                        movs    :offset0a,#palettea
                        movs    :offset1a,#paletteb

                        mov     nibble0,pixels1
                        and     nibble0,#%%3300         'Nibble0 = %%xxxxxxxxxxxxx01xx
                        shr     nibble0,#4              'Nibble0 = %%xxxxxxxxxxxxxxx01
                        add     :offset0a,nibble0

                        mov     nibble2,pixels1
                        and     nibble2,#%%0033         'Nibble2 = %%xxxxxxxxxxxxxxx23
                        add     :offset1a,nibble2

:offset0a               mov     mix1,0-0
:offset1a               or      mix1,0-0

                        movs    :offset0b,#palettea
                        movs    :offset1b,#paletteb

                        mov     nibble0,pixels2
                        and     nibble0,#%%3300         'Nibble0 = %%xxxxxxxxxxxxx45xx
                        shr     nibble0,#4              'Nibble0 = %%xxxxxxxxxxxxxxx45
                        add     :offset0b,nibble0

                        mov     nibble2,pixels2
                        and     nibble2,#%%0033         'Nibble2 = %%xxxxxxxxxxxxxxx67
                        add     :offset1b,nibble2

:offset0b               mov     mix2,0-0
:offset1b               or      mix2,0-0

scan_line_store1        mov     cog_scanline_buffer,mix2
                        add     scan_line_store1,inc_dest_2 'self-modify dest+2

scan_line_store2        mov     cog_scanline_buffer,mix1
                        add     scan_line_store2,inc_dest_2 'self-modify dest+2
                        add     hub_map_ptr,#1          'point to next tile (one instruction between reads)
                        djnz    x,#do_tile   wc         'another tile?

                        ' We've finished rendering the tiles
                        mov     outa,#0
                        ' Verification: if the TV is already asking for this scanline or more
                        ' then panic, go to red alert with the debug LED and skip straight to TV output.
                        ' If the LED does light you need to consider optimising or allocating more
                        ' rendering cogs
                        call    #checktv
        if_ae           mov     outa,#1
        if_ae           jmp     #start_tv_copy

    '                    cmp     current_line,#16 wc,wz
    '    if_ae            jmp     #:wait_for_the_request
'@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
                        ' Render sprites
                        ' find the y-span and reject if current-line isn't in it.
                                                             
                        mov     sprite_attr,rend_sprites

                        mov     sprite_y,rend_sprites
                        and     sprite_y,#$FF
                        mov     draw_y,current_line
                        sub     draw_y,sprite_y  wc,wz
        if_b            jmp     #:wait_for_the_request
                        cmp     draw_y,#16 wc,wz
        if_ae           jmp     #:wait_for_the_request

                        ' find offset in sprite table for this tile
                        mov     tile_number,sprite_attr
                        shr     tile_number,#16
                        and     tile_number,#$FF
                        shl     tile_number,#6          ' 64 bytes in a sprite
                        mov     hub_source,rend_spritetable                               
                        add     hub_source,tile_number

                        ' add offset for y position in the tile
                        add     hub_source,draw_y
                        add     hub_source,draw_y
                        add     hub_source,draw_y
                        add     hub_source,draw_y
 
                        ' put pixel colours in own longs
                        rdlong  pixels,hub_source       'pixels = %%0123456789abcdef

mov     debug_r0,pixels                                                                   '
call    #deb_quart

                        ' initialise self modifiers 
                        movd    :shift_color,#pixel_color_list
                        movd    :shift_mask,#pixel_mask_list
                        movd    :mix_color,#pixel_color_list
                        movd    :mix_mask,#pixel_mask_list
                        ' loop init
                        mov     r0,#4
:outer_loop             mov     count,#4                                       '4*4=16 pixels
:inner_loop             mov     palette_index,pixels                            'get a single pixel
                        and     palette_index,#%%0000000000000003 wc,wz
              if_e      mov     mask,#$FF
              if_ne     mov     mask,#0
                        add     palette_index,#palette                          'and get the adress of it's palette color
                        movs    :lookup_color,palette_index                     '...which is the source of the copy
                        nop    
:lookup_color           mov     color,0-0                  'destination is pixel_color_list
                        ' we now have color and mask
:shift_color            shl     pixel_color_list,#8
:shift_mask             shl     pixel_mask_list,#8
:mix_color              or      pixel_color_list,color
:mix_mask               or      pixel_mask_list,mask
                        shr     pixels,#2
                        ' inner_loop
                        djnz    count,#:inner_loop
                        ' point to next color and mask words
                        add     :shift_color,inc_dest
                        add     :shift_mask,inc_dest
                        add     :mix_color,inc_dest
                        add     :mix_mask,inc_dest
                        ' outer_loop
                        djnz    r0,#:outer_loop
                        ' work out where the sprite starts in the scanline.
                        mov     sprite_x,rend_sprites
                        shr     sprite_x,#10          
                        and     sprite_x,#%00111111
                        mov     cog_dest,#cog_scanline_buffer
                        add     cog_dest,sprite_x
                        'calculate where the sprite stars within a frame
                        mov     sprite_x,rend_sprites
                        shr     sprite_x,#8          
                        and     sprite_x,#%00000011 wc,wz

          '    if_ne     jmp     #:not0
                        ' not shifted

                        mov     shifted4,pixel_color_list
                        mov     shifted3,pixel_color_list+1
                        mov     shifted2,pixel_color_list+2
                        mov     shifted1,pixel_color_list+3
                        mov     shifted5,#0

                        mov     shifted_mask4,pixel_mask_list
                        mov     shifted_mask3,pixel_mask_list+1
                        mov     shifted_mask2,pixel_mask_list+2
                        mov     shifted_mask1,pixel_mask_list+3
                        mov     shifted_mask5,all_background

mov    debug_r0,shifted1
call    #deb_hex
mov     debug_r0,shifted2
call    #deb_hex
mov     debug_r0,shifted3
call    #deb_hex
mov     debug_r0,shifted4
call    #deb_hex
mov     debug_r0,shifted5
call    #deb_hex

mov    debug_r0,shifted_mask1
call    #deb_hex
mov     debug_r0,shifted_mask2
call    #deb_hex
mov     debug_r0,shifted_mask3
call    #deb_hex
mov     debug_r0,shifted_mask4
call    #deb_hex
mov     debug_r0,shifted_mask5
call    #deb_hex


                        jmp     #:end_case
                        
:not0  {                 cmp     sprite_x,#1
              if_ne     jmp     #:not1
                        ' 1 shifted 
                        shl     pixel_color_list,#16
                        shl     pixel_color_list+1,#8
                        
                        shl     pixel_color_list+3,#24
                        shl     pixel_color_list+4,#16
                        shl     pixel_color_list+5,#8
                        
                        shl     pixel_color_list+7,#24
                        shl     pixel_color_list+8,#16
                        shl     pixel_color_list+9,#8
                        
                        shl     pixel_color_list+11,#24
                        shl     pixel_color_list+12,#16
                        shl     pixel_color_list+13,#8

                        shl     pixel_color_list+15,#30

                        mov     shifted1,pixel_color_list
                        or      shifted1,pixel_color_list+1
                        or      shifted1,pixel_color_list+2

                        mov     shifted2,pixel_color_list+3
                        or      shifted2,pixel_color_list+4
                        or      shifted2,pixel_color_list+5
                        or      shifted2,pixel_color_list+6

                        mov     shifted3,pixel_color_list+7
                        or      shifted3,pixel_color_list+8
                        or      shifted3,pixel_color_list+9
                        or      shifted3,pixel_color_list+10

                        mov     shifted4,pixel_color_list+11
                        or      shifted4,pixel_color_list+12
                        or      shifted4,pixel_color_list+13
                        or      shifted4,pixel_color_list+14

                        mov      shifted5,pixel_color_list+15

                        jmp     #:end_case
         }               
:not1     {              cmp     sprite_x,#2
              if_ne     jmp     #:not2
                        '2 shifted
                        shl     pixel_color_list,#8
                        
                        shl     pixel_color_list+2,#24
                        shl     pixel_color_list+3,#16
                        shl     pixel_color_list+4,#8
                        
                        shl     pixel_color_list+6,#24
                        shl     pixel_color_list+7,#16
                        shl     pixel_color_list+8,#8

                        shl     pixel_color_list+10,#24
                        shl     pixel_color_list+11,#16
                        shl     pixel_color_list+12,#8
                        
                        shl     pixel_color_list+14,#24
                        shl     pixel_color_list+15,#16

                        mov     shifted1,pixel_color_list
                        or      shifted1,pixel_color_list+1

                        mov     shifted2,pixel_color_list+2
                        or      shifted2,pixel_color_list+3
                        or      shifted2,pixel_color_list+4
                        or      shifted2,pixel_color_list+5

                        mov     shifted3,pixel_color_list+6
                        or      shifted3,pixel_color_list+7
                        or      shifted3,pixel_color_list+8
                        or      shifted3,pixel_color_list+9

                        mov     shifted4,pixel_color_list+10
                        or      shifted4,pixel_color_list+11
                        or      shifted4,pixel_color_list+12
                        or      shifted4,pixel_color_list+13

                        mov      shifted5,pixel_color_list+14
                        or     shifted5,pixel_color_list+15
                       jmp     #:end_case
           }
:not2      {             ' must be 3 shifted
                        shl     pixel_color_list+1,#24
                        shl     pixel_color_list+2,#16
                        shl     pixel_color_list+3,#8
                        
                        shl     pixel_color_list+5,#24
                        shl     pixel_color_list+6,#16
                        shl     pixel_color_list+7,#8
                        
                        shl     pixel_color_list+9,#24
                        shl     pixel_color_list+10,#16
                        shl     pixel_color_list+11,#8
                        
                        shl     pixel_color_list+12,#24
                        shl     pixel_color_list+13,#16
                        shl     pixel_color_list+14,#8

                        shl     pixel_color_list+15,#24

                        mov     shifted1,pixel_color_list
                        mov     shifted2,pixel_color_list+1
                        or      shifted2,pixel_color_list+2
                        or      shifted2,pixel_color_list+3
                        or      shifted2,pixel_color_list+4
                        mov     shifted3,pixel_color_list+5
                        or      shifted3,pixel_color_list+6
                        or      shifted3,pixel_color_list+7
                        or      shifted3,pixel_color_list+8
                        mov     shifted4,pixel_color_list+9
                        or      shifted4,pixel_color_list+10
                        or      shifted4,pixel_color_list+11
                        or      shifted4,pixel_color_list+12
                        mov     shifted4,pixel_color_list+13
                        or      shifted4,pixel_color_list+14
                        or      shifted5,pixel_color_list+15

                        ' calculate mask
                        'mov     mask1,shifted1
                        
        }
:end_case               movd    :write_it,cog_dest
                        movs    :write_it,#shifted1
                        movd    :mask_it,cog_dest
                        movs    :mask_it,#shifted_mask1
                        mov     count,#5
:mask_it                and     0-0,0-0            
:write_it               or      0-0,0-0
                        add     :write_it,inc_dest
                        add     :write_it,#1
                        add     :mask_it,inc_dest
                        add     :mask_it,#1
                        djnz    count,#:mask_it
    
'@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

                        ' We've finished rendering the line and we didn't have a hissy fit panic.
                        ' We need to wait here for the line we just prepared to be requested by the TV driver
:wait_for_the_request
                        call    #checktv
        if_b            jmp     #:wait_for_the_request

                        ' Our line has been requested so copy it to hub scan line buffer
start_tv_copy           movs    :nextcopy, #cog_scanline_buffer
                        mov     ptr, rend_scanline_buffer
                        mov     count, #64              ' length of visible scanline buffer in longs
:nextcopy               mov     pixels,cog_scanline_buffer
                        add     :nextcopy, #1
                        wrlong  pixels, ptr
                        add     ptr, #4
                        djnz    count, #:nextcopy

                        ' We've handed over the scanline to the TV driver, so do another.
allocate_next_job       add     current_line,rend_cog_count 'we only do every Nth line
                        ' Degrade gracefully when there's too much work to do.
                        call    #checktv
              if_ae     jmp     #allocate_next_job
                        ' Wrap around.
                        cmp     current_line,#224 wc
              if_ae     sub     current_line,#224
                        jmp     #do_line


'-------------------------------------------------------------------
checktv                 ' Tf we're currently rendering a low numbered line when a high one is being requested
                        ' then that's OK because we wrapped around.



                        rdlong  scanline_req,rend_scanline_req_adr
                        cmp     scanline_req,#210 wc,wz
        if_be           jmp     #:checkit
                        cmp     current_line,#7 wc,wz
        if_be           cmp     scanline_req, current_line wz
                        'b = normal
                        'z = doit
                        'a = err
        if_be           jmp     #checktv_ret

:checkit                cmp     scanline_req, current_line wz, wc    'set carry if request is less than current.
                        'b = normal
                        'z = doit
                        'a = err

                        ' At this point, we could skip drawing sprites or something like that and go
                        ' straight to image output and still stay synced with the tv driver
                        ' We must be quick though!
                        'jmp #start_tv_copy
checktv_ret             ret
'-------------------------------------------------------------------

'Constants
sprite_is_visible       long    $01000000
draw_y                  long    0

sprite_x                long    0
sprite_y                long    0

pixels                  long    0
sprite_attr             long    0

'
' Initialized data
'
inc_dest                long    1 << 9 << 0
inc_dest_2              long    1 << 9 << 1

'
' Uninitialized data
'
x                       long    0                       'display
tile                    long    0
pixels1                 long    0
pixels2                 long    0
pixels3                 long    0
pixels4                 long    0
'mix1                    long    0
'mix2                    long    0
'mix3                    long    0
'mix4                    long    0
'mix5                    long    0
'mix6                    long    0
'mix7                    long    0
'mix8                    long    0
mix1
shifted1                long    0
mix2
shifted2                long    0
shifted3                long    0
shifted4                long    0
shifted5                long    0
r0                      long    0
r1                      long    0
r2                      long    0
tile_number             long    0
hub_source              long    0
cog_dest                long    0
mask                    long    0
color                   long    0

pixel_color_list        long    0
                        long    0
                        long    0
                        long    0

pixel_mask_list         long    0
                        long    0
                        long    0
                        long    0

shifted_mask1           long    0
shifted_mask2                        long    0
shifted_mask3                        long    0
shifted_mask4                        long    0      
shifted_mask5           long    0
all_background          long    $FFFFFFFF

DAT  ' new stuff

ptr                     long    0
count                   long    0
nibble0                 long    0
nibble2                 long    0


hub_map_ptr             long    0
current_line            long    0
tile_table              long    0
tile_line               long    0
tile_row                long    0

' Parameter buffer
renderer_params
rend_scanline_req_adr   long    0
rend_map                long    0
rend_tiletable          long    0
rend_spritetable        long    0
                        long    0               'rend_colors
rend_ht                 long    0               '
rend_vt                 long    0
rend_scanline_buffer    long    0
rend_cog_number         long    0
rend_cog_count          long    0

rend_sprites            long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0
                        long    0

rend_debugging          long    0


' End of parameter buffer

scanline_req
                        long    1

' Cog scanline buffer = 256 pixels, 8 bits per pixel = 2048 bits = 256 bytes = 64 longs
' However, sprites are going to be 16x16 and I want the facility to be able to slide sprites off
' screen to the left or right without explicit clipping code or corrupting adjacent cog_memory
' So an extra 16 pixels = 128 bits = 4 longs either side of the visible scan buffer
                        long    0
                        long    0
                        long    0
                        long    0
cog_scanline_buffer
colorbar0               long    $0b_0b_0b_0b
colorbar1               long    $0c_0c_0c_0c
colorbar2               long    $0d_0d_0d_0d
colorbar3               long    $0e_0e_0e_0e
colorbar4               long    $1b_0b_1b_1b
colorbar5               long    $1c_0c_1c_1c
colorbar6               long    $1d_0d_1d_1d
colorbar7               long    $1e_0e_1e_1e
colorbar8               long    $2b_0b_2b_2b
colorbar9               long    $2c_2c_2c_2c
colorbar10              long    $2d_2d_2d_2d
colorbar11              long    $2e_2e_2e_2e
colorbar12              long    $3b_3b_3b_3b
colorbar13              long    $3c_3c_3c_3c
colorbar14              long    $3d_3d_3d_3d
colorbar15              long    $3e_3e_3e_3e
colorbar16              long    $4b_4b_4b_4b
colorbar17              long    $4c_4c_4c_4c
colorbar18              long    $4d_4d_4d_4d
colorbar19              long    $4e_4e_4e_4e
colorbar20              long    $5b_5b_5b_5b
colorbar21              long    $5c_5c_5c_5c
colorbar22              long    $5d_5d_5d_5d
colorbar23              long    $5e_5e_5e_5e
colorbar24              long    $6b_6b_6b_6b
colorbar25              long    $6c_6c_6c_6c
colorbar26              long    $6d_6d_6d_6d
colorbar27              long    $6e_6e_6e_6e
colorbar28              long    $7b_7b_7b_7b
colorbar29              long    $7b_7b_7b_7b
colorbar30              long    $7b_7b_7b_7b
colorbar31              long    $7b_7b_7b_7b
colorbar32              long    $7b_7b_7b_7b
colorbar33              long    $7b_7b_7b_7b
colorbar34              long    $7b_7b_7d_7b
colorbar35              long    $7b_7b_7b_7b
colorbar36              long    $9b_9b_9b_9b
colorbar37              long    $9c_9c_9c_9c
colorbar38              long    $9d_9d_9d_9d
colorbar39              long    $9e_9e_9e_9e
colorbar40              long    $ab_ab_ab_ab
colorbar41              long    $ac_ac_ac_ac
colorbar42              long    $ad_ad_ad_ad
colorbar43              long    $ae_ae_ae_ae
colorbar44              long    $bb_bb_bb_bb
colorbar45              long    $bc_bc_bc_bc
colorbar46              long    $bd_bd_bd_bd
colorbar47              long    $be_be_be_be
colorbar48              long    $cb_cb_cb_cb
colorbar49              long    $cc_cc_cc_cc
colorbar50              long    $cd_cd_cd_cd
colorbar51              long    $ce_ce_ce_ce
colorbar52              long    $db_db_db_db
colorbar53              long    $dc_dc_dc_dc
colorbar54              long    $dd_dd_dd_dd
colorbar55              long    $de_de_de_de
colorbar56              long    $eb_eb_eb_eb
colorbar57              long    $ec_ec_ec_ec
colorbar58              long    $ed_ed_ed_ed
colorbar59              long    $ee_ee_ee_ee
colorbar60              long    $fb_fb_fb_fb
colorbar61              long    $fc_fc_fc_fc
colorbar62              long    $fd_fd_fd_fd
colorbar63              long    $fe_fe_fe_fe
' End of scanline buffer
                        long    0
                        long    0
                        long    0
                        long    0

'tiles
palettea
                        long    $52520000
                        long    $52dd0000
                        long    $52050000
                        long    $525b0000

                        long    $dd520000
                        long    $dddd0000
                        long    $dd050000
                        long    $dd5b0000

                        long    $05520000
                        long    $05dd0000
                        long    $05050000
                        long    $055b0000

                        long    $5b520000
                        long    $5bdd0000
                        long    $5b050000
                        long    $5b5b0000

paletteb
                        long    $00005252
                        long    $000052dd
                        long    $00005205
                        long    $0000525b

                        long    $0000dd52
                        long    $0000dddd
                        long    $0000dd05
                        long    $0000dd5b

                        long    $00000552
                        long    $000005dd
                        long    $00000505
                        long    $0000055b

                        long    $00005b52
                        long    $00005bdd
                        long    $00005b05
                        long    $00005b5b

'sprites
                       ' 52 = red hued black
palette_index           long    0

palette                 long    $00000000
                        long    $0000000b
                        long    $0000005b
                        long    $00000006

VAR
  byte  debug_map[28*32]
  long  row
  long  col
  
PUB debug_init(map)
  bytefill(@debug_map,151,32*28)
  long[0][map/4] := @debug_map
  
PUB debug(map)
  long[0][map/4] := @debug_map

PUB _debug_hex(number) | adr
  
  debug_map[row*32+col++] := number >> 28   
  debug_map[row*32+col++] := (number >> 24) & $F
  debug_map[row*32+col++] := (number >> 20) & $F
  debug_map[row*32+col++] := (number >> 16) & $F
  debug_map[row*32+col++] := (number >> 12) & $F
  debug_map[row*32+col++] := (number >> 8) & $F
  debug_map[row*32+col++] := (number >> 4) & $F
  debug_map[row*32+col++] := number & $F

PUB debug_new_line
  row++
  col:=0

DAT
debug_col               long    0
debug_row               long    0
debug_hexchar           long    170
debug_quartchar         long    70
debug_r0                long    0
debug_r1                long    0
debug_count             long    0
debug_ptr               long    0
'----------------------------------------------------------------
debug_start
                        mov     debug_col,#0
                        mov     debug_row,#4
debug_start_ret         ret
'----------------------------------------------------------------
deb_hex
                        mov     debug_ptr,PAR
                        add     debug_ptr,#param_count*4+32*4 
                        rdlong  rend_debugging,debug_ptr wc,wz
        if_e            jmp     #deb_hex_ret
                        add     rend_debugging,debug_col
                        mov     debug_count,debug_row wc,wz
        if_e            jmp     #:doit
:loop1                  add     rend_debugging,#32
                        djnz    debug_count,#:loop1

:doit                   wrbyte  debug_hexchar,rend_debugging
                        add     rend_debugging,#1
                        mov     debug_count,#8
:loop2                  mov     debug_r1,debug_r0
                        shr     debug_r1,#28                        
                        wrbyte  debug_r1,rend_debugging
                        add     rend_debugging,#1
                        shl     debug_r0,#4
                        djnz    debug_count,#:loop2
                        add     debug_row,#1
deb_hex_ret             ret
'---------------------------------------------------------------
deb_quart
                        'get map
                        mov     debug_ptr,PAR
                        add     debug_ptr,#param_count*4+32*4 
                        rdlong  rend_debugging,debug_ptr wc,wz
        if_e            jmp     #deb_quart_ret
                        'move to cursor pos
                        add     rend_debugging,debug_col
                        mov     debug_count,debug_row wc,wz
        if_e            jmp     #:doit
:loop1                  add     rend_debugging,#32
                        djnz    debug_count,#:loop1
                        'output base marker
:doit                   wrbyte  debug_quartchar,rend_debugging
                        add     rend_debugging,#1
                        'loop round digits
                        mov     debug_count,#16
:loop2                  mov     debug_r1,debug_r0
                        shr     debug_r1,#30
                        wrbyte  debug_r1,rend_debugging
                        add     rend_debugging,#1
                        shl     debug_r0,#2
                        djnz    debug_count,#:loop2
                        add     debug_row,#1
deb_quart_ret           ret
'----------------------------------------------------------------

                        fit