' /////////////////////////////////////////////////////////////////////////////
' Donkey Kong Remake
' AUTHOR: Steve Waddicor
'//////////////////////////////////////////////////////////////////////////////

CON

  _clkmode = xtal1 + pll8x
  _xinfreq = 10_000_000 + 0000
  _stack = 100 'accomodate display memory and stack

  tv_paramcount = 17
  cogs = 2

VAR
  ' Renderer parameter block
  ' Initializing data is in the DAT block at the end of this file.
  ' This is copied in to this variable space with a longmove
  ' so it can be altered easily (without remembering to use @ all the time).
  ' Each renderer then copies these values to it's own cog_memory cache
  ' when it's got a little spare time, several times per frame
  ' so these values can be altered on the fly. 
  long  rend_scanline_req_adr    'a pointer to the request to put a line in the scan buffer   write-only
  long  rend_map                 'pointer to screen (bytes)                                write-only
  long  rend_tiletable           'pointer to tile definitions (words)                      write-only
  long  rend_spritetable         'pointer to tile definitions (longs)                      write-only
  long  rend_colors              'pointer to colors (longs)                                write-only               
  long  rend_ht                  'horizontal tiles                                         write-only
  long  rend_vt                  'vertical tiles                                           write-only
  long  rend_scanline_buffer_adr 'hug memory where renderer puts scanline when requested   write-only
  long  rend_cog_number          'each cog gets a different number, starting with zero     write-only
  long  rend_cog_count           'total number of cogs                                     write-only
  long  rend_sprites[32]         '%_______________mfv_xxxxxxxx_yyyyyyyy
  long  debug_map
  byte  startkey
  byte  debug

OBJ

  tv    : "sw_dk_tv_drv_018.spin"
  gfx   : "sw_dkong_gfx_002.spin"
  rend  : "sw_dk_gfx_renderer_007.spin"

PUB start | i, gp


    repeat i from 0 to constant(cogs-1)
        ' start a renderer
        rend_tiletable := gfx.tile_adr
        rend_spritetable := gfx.sprite_adr
        rend_map := gfx.map_adr
        rend_ht := 32
        rend_vt := 28
        rend_scanline_buffer_adr := @hub_scanline_buffer
        rend_scanline_req_adr := @scanline_req
        rend_cog_number := i
        rend_cog_count := cogs
        rend_sprites := $010030C8
        rend.start(@rend_scanline_req_adr)
        repeat 100000
     
    ' start the TV driver 
    tv.start(@tv_params)

    rend.debug_init(@debug_map)
    debug_map:=0
  
    repeat
        gp := NES_Read_Gamepad

        if gp & %1000
            rend_sprites.BYTE[0]-=1
            rend_sprites.BYTE[2]+=1            
            if rend_sprites.BYTE[2]>4
               rend_sprites.BYTE[2]:=3 
        elseif gp & %0100
            rend_sprites.BYTE[0]+=1
            rend_sprites.BYTE[2]-=1        
            if rend_sprites.BYTE[2]<3
               rend_sprites.BYTE[2]:=4 
        elseif gp & %0010
            rend_sprites.BYTE[1]-=4
            rend_sprites.BYTE[2]-=1            
            if rend_sprites.BYTE[2]==255
               rend_sprites.BYTE[2]:=2 
        elseif gp & %0001
            rend_sprites.BYTE[1]+=4
            rend_sprites.BYTE[2]+=1        
            if rend_sprites.BYTE[2]>2
               rend_sprites.BYTE[2]:=0 

        if gp & %00010000
            if startkey==0
                startkey:=1
                !debug
                if debug
                    rend.debug(@debug_map)
                else
                    debug_map:=0
        else
            startkey:=0

        repeat 30000
      

PUB NES_Read_Gamepad : nes_bits        |  i

' //////////////////////////////////////////////////////////////////
' NES Game Paddle Read
' //////////////////////////////////////////////////////////////////       
' reads both gamepads in parallel encodes 8-bits for each in format
' right game pad #1 [15..8] : left game pad #0 [7..0]
'
' set I/O ports to proper direction
' P3 = JOY_CLK      (4)
' P4 = JOY_SH/LDn   (5) 
' P5 = JOY_DATAOUT0 (6)
' P6 = JOY_DATAOUT1 (7)
' NES Bit Encoding
'
' RIGHT  = %00000001
' LEFT   = %00000010
' DOWN   = %00000100
' UP     = %00001000
' START  = %00010000
' SELECT = %00100000
' B      = %01000000
' A      = %10000000

' step 1: set I/Os
DIRA [3] := 1 ' output
DIRA [4] := 1 ' output
DIRA [5] := 0 ' input
DIRA [6] := 0 ' input

' step 2: set clock and latch to 0
OUTA [3] := 0 ' JOY_CLK = 0
OUTA [4] := 0 ' JOY_SH/LDn = 0
'Delay(1)

' step 3: set latch to 1
OUTA [4] := 1 ' JOY_SH/LDn = 1
'Delay(1)

' step 4: set latch to 0
OUTA [4] := 0 ' JOY_SH/LDn = 0

' step 5: read first bit of each game pad

' data is now ready to shift out
' first bit is ready 
nes_bits := 0

' left controller
nes_bits := INA[5] | (INA[6] << 8)

' step 7: read next 7 bits
repeat i from 0 to 6
 OUTA [3] := 1 ' JOY_CLK = 1
 'Delay(1)             
 OUTA [3] := 0 ' JOY_CLK = 0
 nes_bits := (nes_bits << 1)
 nes_bits := nes_bits | INA[5] | (INA[6] << 8)

 'Delay(1)             
' invert bits to make positive logic
nes_bits := (!nes_bits & $FFFF)

' //////////////////////////////////////////////////////////////////
' End NES Game Paddle Read
' //////////////////////////////////////////////////////////////////  

DAT

scanline_req            long    0

tv_params               long    @scanline_req+$10            'status
                        long    @hub_scanline_buffer+$10     'enable

' Hub scanline buffer = 256 pixels, 8 bits per pixel = 2048 bits = 256 bytes = 64 longs
' However, sprites are going to be 16x16 and I want the facility to be able to slide sprites off
' screen to the left or right without explicit clipping code or corrupting adjacent cog_memory
' So an extra 16 pixels = 128 bits = 4 longs either side of the visible scan buffer
                        long     0
                        long     0
                        long     0
                        long     0                        
hub_scanline_buffer
colorbar0               long    $0b_0b_0b_0b
colorbar1               long    $0c_0c_0c_0c
colorbar2               long    $0d_0d_0d_0d
colorbar3               long    $0e_0e_0e_0e
colorbar4               long    $1b_0b_1b_1b
colorbar5               long    $1c_0c_1c_1c
colorbar6               long    $1d_0d_1d_1d
colorbar7               long    $1e_0e_1e_1e
colorbar8               long    $2b_0b_2b_2b
colorbar9               long    $2c_2c_2c_2c
colorbar10              long    $2d_2d_2d_2d
colorbar11              long    $2e_2e_2e_2e
colorbar12              long    $3b_3b_3b_3b
colorbar13              long    $3c_3c_3c_3c
colorbar14              long    $3d_3d_3d_3d
colorbar15              long    $3e_3e_3e_3e
colorbar16              long    $4b_4b_4b_4b
colorbar17              long    $4c_4c_4c_4c
colorbar18              long    $4d_4d_4d_4d
colorbar19              long    $4e_4e_4e_4e
colorbar20              long    $5b_5b_5b_5b
colorbar21              long    $5c_5c_5c_5c
colorbar22              long    $5d_5d_5d_5d
colorbar23              long    $5e_5e_5e_5e
colorbar24              long    $6b_6b_6b_6b
colorbar25              long    $6c_6c_6c_6c
colorbar26              long    $6d_6d_6d_6d
colorbar27              long    $6e_6e_6e_6e
colorbar28              long    $7b_7b_7b_7b
colorbar29              long    $7c_7c_7c_7c
colorbar30              long    $7d_7d_7d_7d
colorbar31              long    $7e_7e_7e_7e
colorbar32              long    $8b_8b_8b_8b
colorbar33              long    $8c_8c_8c_8c
colorbar34              long    $8d_8d_8d_8d
colorbar35              long    $8e_8e_8e_8e
colorbar36              long    $9b_9b_9b_9b
colorbar37              long    $9c_9c_9c_9c
colorbar38              long    $9d_9d_9d_9d
colorbar39              long    $9e_9e_9e_9e
colorbar40              long    $ab_ab_ab_ab
colorbar41              long    $ac_ac_ac_ac
colorbar42              long    $ad_ad_ad_ad
colorbar43              long    $ae_ae_ae_ae
colorbar44              long    $bb_bb_bb_bb
colorbar45              long    $bc_bc_bc_bc
colorbar46              long    $bd_bd_bd_bd
colorbar47              long    $be_be_be_be
colorbar48              long    $cb_cb_cb_cb
colorbar49              long    $cc_cc_cc_cc
colorbar50              long    $cd_cd_cd_cd
colorbar51              long    $ce_ce_ce_ce
colorbar52              long    $db_db_db_db
colorbar53              long    $dc_dc_dc_dc
colorbar54              long    $dd_dd_dd_dd
colorbar55              long    $de_de_de_de
colorbar56              long    $eb_eb_eb_eb
colorbar57              long    $ec_ec_ec_ec
colorbar58              long    $ed_ed_ed_ed
colorbar59              long    $ee_ee_ee_ee
colorbar60              long    $fb_fb_fb_fb
colorbar61              long    $fc_fc_fc_fc
colorbar62              long    $fd_fd_fd_fd
colorbar63              long    $fe_fe_fe_fe

                        long    0
                        long    0
                        long    0
                        long    0
' End of scanline buffer